(* tyconstr_id is the name of an datatype constructor.
   label_id is the name of a field label or data constructor. *)
type tyconstr_id = Type of string
type label_id = Label of string

type kind = Variant | Record

type ('b,'t) decl = {
    name : tyconstr_id ;
    type_params : 'b list ;
    data_kind : kind ;
    labels_descr : 't label list ;
  }

and 't label = {
    label_name : label_id ;
    type_name : tyconstr_id ;
    arg_type : 't ;
  }

(* During inference, we will need an environment to keep track of
 * the user defined types *)

module LabelId = struct
  type t = label_id
  let compare = Stdlib.compare
end

module LabelMap =
  Map.Make(LabelId)

module TyConstrId = struct
  type t = tyconstr_id
  let compare = Stdlib.compare
end

module TyConstrMap =
  Map.Make(TyConstrId)

module Env = struct

  (* Environment for user defined data types *)
  type ('b,'t) t = {
      datatypes :  ('b,'t) decl TyConstrMap.t;
      labels : 't label LabelMap.t;
    }

  exception UnexpectedRecord

  let empty = {
      datatypes = TyConstrMap.empty;
      labels = LabelMap.empty;
    }

  let add_decl e tdescr =
    let add_datatype m tdescr =
      TyConstrMap.add tdescr.name tdescr m in
    let add_labels m ldescrs =
      List.fold_left (fun m ldescr ->
          LabelMap.add ldescr.label_name ldescr m
        ) m ldescrs
    in {
      datatypes = add_datatype e.datatypes tdescr;
      labels = add_labels e.labels tdescr.labels_descr;
    }

  let find_label { labels ; _ } (Label l) =
    LabelMap.find (Label l) labels

  let find_decl { datatypes ; _ } (Type tid) =
    TyConstrMap.find (Type tid) datatypes

  let label_index { labels_descr ; _ } l =
    let combine i ldescr = (i, ldescr.label_name) in
    List.mapi combine labels_descr
    |> List.find (fun (_, l') -> l = l')
    |> fst

  let () = ignore label_index

  let map (type b1 b2 t1 t2)
    (f : (b1, t1) decl -> (b2, t2) decl)
    (env : (b1, t1) t) : (b2, t2) t
  =
    TyConstrMap.to_seq env.datatypes
    |> Seq.map snd
    |> Seq.map f
    |> Seq.fold_left add_decl empty

end
