open Client

(* -------------------------------------------------------------------------- *)

(* A random generator of pure lambda-terms. *)

let int2var k =
  "x" ^ string_of_int k

(* [split n] produces two numbers [n1] and [n2] comprised between [0] and [n]
   (inclusive) whose sum is [n]. *)

let split n st =
  let n1 = Random.State.int st (n + 1) in
  let n2 = n - n1 in
  n1, n2

let rec size accu = function
  | ML.Var _
  | ML.Variant (_, _, None)
    -> accu
  | ML.Abs (_, _, t)
  | ML.Variant (_, _, Some t)
  | ML.Annot (_, t, _)
    -> size (accu + 1) t
  | ML.App (_, t1, t2)
  | ML.Let (_, _, t1, t2)
  | ML.LetProd (_, _, t1, t2)
    -> size (size (accu + 1) t1) t2
  | ML.Tuple (_, ts)
  | ML.Hole (_, ts)
    -> List.fold_left size (accu + 1) ts
  | ML.Match (_, t, brs)
    -> List.fold_left size_branch (size (accu + 1) t) brs

and size_branch accu (pat,u) =
  size (size_pattern accu pat) u

and size_pattern accu = function
  | ML.PVar _
  | ML.PVariant (_, _, None)
    -> accu
  | ML.PWildcard _
    -> accu
  | ML.PTuple (_, ps)
    -> List.fold_left size_pattern (accu + 1) ps
  | ML.PVariant (_, _, Some p)
  | ML.PAnnot (_, p, _)
    -> size_pattern (accu + 1) p


let size =
  size 0


(* -------------------------------------------------------------------------- *)

(* A random term generator using QCheck. *)

let dummy_pos = ML.dummy_pos

let var x = ML.Var (dummy_pos, (int2var x))

(* In the following functions, the parameter [k] is the number of free
   variables; the parameter [n] is the size (i.e., the number of
   internal nodes). *)

let abs k n self = QCheck.Gen.(
    let+ (x,t) = pair (int2var <$> pure k) (self (k + 1, n - 1))
    in ML.Abs (dummy_pos, x, t)
  )

let app k n self = QCheck.Gen.(
    let* n1, n2 = split (n - 1) in
    let+ (t1, t2) = pair (self (k, n1)) (self (k, n2))
    in ML.App (dummy_pos, t1, t2)
  )

let tuple k n self = QCheck.Gen.(
    let* n1, n2 = split (n - 1) in
    let+ (t1, t2) = pair (self (k, n1)) (self (k, n2))
    in ML.Tuple (dummy_pos, [t1; t2])
  )

let let_ k n self = QCheck.Gen.(
    let* n1, n2 = split (n - 1) in
    let+ (x,t1,t2) =
      triple (int2var <$> pure k) (self (k, n1)) (self (k+1, n2))
    in ML.Let (dummy_pos, x, t1, t2)
  )

let let_prod k n self = QCheck.Gen.(
    let* n1, n2 = split (n - 1) in
    let+ (x1,x2,t1,t2) =
      quad
        (int2var <$> pure k)
        (int2var <$> pure (k + 1))
        (self (k, n1))
        (self (k+2, n2))
    in ML.LetProd (dummy_pos, [x1; x2], t1, t2)
  )

let annot k n self = QCheck.Gen.(
  let+ t = self (k, n-1) in
  ML.Annot (dummy_pos, t, (["'a"], ML.TyVar (dummy_pos, "'a")))
)

let term_gen = QCheck.Gen.(
    fix
      (fun self (k, n) ->
        if n = 0 then begin
          assert (k > 0);
          var <$> (int_bound (k-1))
        end
        else
          if k = 0 then
            abs k n self
          else
           frequency
             [
               1, abs k n self ;
               1, app k n self ;
               1, tuple k n self ;
               1, let_ k n self ;
               1, let_prod k n self ;
               1, annot k n self ;
             ]
      )
  )

(* -------------------------------------------------------------------------- *)

(* Shrinker. *)

module Shrinker = struct
  module Shrink = QCheck.Shrink
  module Iter = QCheck.Iter

  let rec bv_pat = function
    | ML.PVar (_, y) -> [y]
    | ML.PWildcard _ | ML.PVariant (_, _, None) -> []
    | ML.PAnnot (_, pat, _)
    | ML.PVariant (_, _, Some pat) ->
       bv_pat pat
    | ML.PTuple (_, ps) ->
       List.(concat (map bv_pat ps))

  (* t[u/x] *)
  let rec subst t x u =
    match t with
    | ML.Var (pos, y) ->
      if x = y then u
      else ML.Var (pos, y)
    | ML.Abs (pos, y, t) ->
      ML.Abs (pos, y, subst_under [y] t x u)
    | ML.App (pos, t1, t2) ->
      ML.App (pos, subst t1 x u, subst t2 x u)
    | ML.Let (pos, y, t1, t2) ->
      ML.Let (pos, y, subst t1 x u, subst_under [y] t2 x u)
    | ML.Tuple (pos, ts) ->
      ML.Tuple (pos, List.map (fun t -> subst t x u) ts)
    | ML.Hole (pos, ts) ->
      ML.Hole (pos, List.map (fun t -> subst t x u) ts)
    | ML.LetProd (pos, ys, t1, t2) ->
      ML.LetProd (pos, ys, subst t1 x u, subst_under ys t2 x u)
    | ML.Annot (pos, t, ty) ->
      ML.Annot (pos, subst t x u, ty)
    | ML.Variant (pos, ty, t) ->
      ML.Variant (pos, ty, Option.map (fun t -> subst t x u) t)
    | ML.Match (pos, t, brs) ->
      ML.Match (pos, subst t x u, List.map (fun br -> subst_branch br x u) brs)

  (* (y.t)[u/x] *)
  and subst_under ys t x u =
    if List.mem x ys then t
    else subst t x u

  and subst_under_pat p t x u =
    subst_under (bv_pat p) t x u

  and subst_branch (p, t) x u =
    (p, subst_under_pat p t x u)

  let remove_variable t x =
    subst t x (ML.Hole (dummy_pos, []))

  let remove_variables t xs =
    List.fold_left remove_variable t xs

  let rec shrink_term t = Iter.(
    (* The Iter product is not suitable for shrinking:
         (and+) (shrink a) (shrink b)
       will try to shrink *both* a and b simultaneously,
       but never try to shrink just one of them.

       We shadow this unsuitable operator and use a "pointed product" (and++),
       which tries to shrink a (with b constant) and then b (with a constant).

       This corresponds to the behavior of Shrink.pair, but Shrink.pair works
       on function types ('a -> 'a Iter.t), which is inconvenient
       to write shrinkers directly -- no 'map' function. *)
    let (and+) = () in ignore ((and+));
    let (let++) (_a, it) f = Iter.map f it in
    let (and++) (a, ita) (b, itb) =
      (a, b),
      Iter.append
        (let+ a' = ita in (a', b))
        (let+ b' = itb in (a, b'))
    in
    let subterms ts =
      (* ...[t1, t2] is not always a desirable shrinking choice for a term
         containing [t1, t2] as subterms, because it breaks the property that
         the types of t1 and t2 appear in the resulting type, which is sometimes
         important to reproduce typing bugs.
         So we start with just (t1) and (t2) as shrinking choices,
         even though they typically do not preserve typability. *)
      of_list ts <+> return (ML.Hole (dummy_pos, ts)) in
    (match t with
     | ML.Hole (_, []) ->
        empty
     | _ ->
       return (ML.Hole (dummy_pos, []))
    ) <+> (
      match t with
      | ML.Var _ ->
         empty

      | ML.Hole (pos, ts) ->
        (match ts with [t] -> return t | _ -> empty)
        <+> (
         let+ ts' = Shrink.list ~shrink:shrink_term ts
         in ML.Hole (pos, ts')
       )

      | ML.Abs (pos, x, t) ->
         subterms [remove_variable t x]
         <+> (
           let+ t' = shrink_term t in
           ML.Abs (pos, x, t')
         )

      | ML.App (pos, t1, t2) ->
        subterms [t1; t2]
        <+> (
         let++ t1' = t1, shrink_term t1
         and++ t2' = t2, shrink_term t2
         in ML.App (pos, t1',t2')
       )

      | ML.Let (pos, x, t, u) ->
        subterms [t; remove_variable u x]
        <+> (
         let++ t' = t, shrink_term t
         and++ u' = u, shrink_term u
         in ML.Let (pos, x, t', u')
       )

      | ML.LetProd (pos, xs, t, u) ->
        subterms [t; remove_variables u xs]
        <+> (match t with
          | ML.Hole _ ->
            let+ xs' = Shrink.list_spine xs in
            let u =
              List.filter (fun x -> not (List.mem x xs')) xs
              |> remove_variables u
            in
            if xs' = [] then u
            else ML.LetProd (pos, xs', t, u)
          | _ -> empty
        )
        <+> (
         let++ t' = t, shrink_term t
         and++ u' = u, shrink_term u
         in ML.LetProd (pos, xs, t', u')
       )

      | ML.Tuple (pos, ts) ->
        subterms ts
         <+> (
           let+ ts' = Shrink.list ~shrink:shrink_term ts
           in ML.Tuple (pos, ts')
         )

      | ML.Annot (_, t, _ty) ->
        subterms [t]

      | ML.Variant (pos, lab, t) ->
        subterms (Option.to_list t)
        <+> (
          match t with
          | None -> empty
          | Some t ->
            let+ t' = shrink_term t
            in ML.Variant (pos, lab, Some t')
        )

      | ML.Match (_, _t, _branches) ->
        failwith "unsupported"
    )
  )
end

(* -------------------------------------------------------------------------- *)

(* Random testing. *)

(* A list of pairs [m, n], where [m] is the number of tests and [n] is the
   size of the randomly generated terms. *)

let pairs =
  match Sys.getenv_opt "INFERNO_SLOW_TESTS" with
  | Some ("0" | "false" | "") | None -> [
      10_000, 10;
      10_000, 15;
    ]
  | Some _ -> [
      100_000, 5;
      100_000, 10;
      100_000, 15;
      100_000, 20;
      100_000, 25; (* at this size, about 1% of the terms are well-typed *)
      100_000, 30;
      (* At the following sizes, no terms are well-typed! *)
      10_000, 100;
      10_000, 500;
      1000, 1000;
      100, 10_000;
      10, 100_000;
      1, 1_000_000;
    ]

let () =
  Printf.printf "Preparing to type-check a bunch of randomly-generated terms...\n%!";
  let count_success = ref 0 in
  let count_total = ref 0 in
  let tests =
    List.map (fun (m, n) ->
      let arb =
        QCheck.make
          ~print:MLPrinter.to_string
          ~shrink:Shrinker.shrink_term
          (term_gen (0, n))
      in
      let prop t =
        Test.CheckML.test_with_args
          ~success:count_success
          ~total:count_total
          Datatype.Env.empty
          t
      in
      QCheck.Test.make
        ~name:(Printf.sprintf "Type checking (size %d)" n)
        ~count:m
        arb
        prop
  ) pairs in

  (* Note that we need to initialize the random state on each test. *)
  let suite () =
    QCheck_runner.set_seed 0;
    let outcome = QCheck_runner.run_tests
                    ~debug_shrink:(Some stdout)
                    tests
    in
    if outcome <> 0 then
      failwith "Problem detected !"
    else begin
        Printf.printf
          "\nIn total, %d out of %d terms (%.1f%%) were considered well-typed.\n%!"
          !count_success !count_total
          (float !count_success /. float !count_total *. 100.);
        Printf.printf "No problem detected.\n%!"
      end
  in
  Alcotest.run "random suite" [ ("random suite",
                                 [ Alcotest.test_case "random suite" `Quick suite ] ) ]
    ~and_exit:false ~verbose:false ;
