(******************************************************************************)
(*                                                                            *)
(*                                  Inferno                                   *)
(*                                                                            *)
(*                       François Pottier, Inria Paris                        *)
(*                                                                            *)
(*  Copyright Inria. All rights reserved. This file is distributed under the  *)
(*  terms of the MIT License, as described in the file LICENSE.               *)
(*                                                                            *)
(******************************************************************************)

open UnifierSig
open SolverSig

(* This module offers the traditional, low-level presentation of a constraint
   solver: a definition of the syntax of constraints and a function [solve],
   which solves a constraint and annotates it, so as to publish information
   that can be exploited in the reconstruction phase. *)

module Make
  (X : TEVAR)
  (S : STRUCTURE)
  (O : OUTPUT with type 'a structure = 'a S.structure)
: sig

  (* The type [tevar] of term variables is provided by [X]. *)
  open X

  (* The type ['a structure] of shallow types is provided by [S]
     (and repeated by [O]). *)

  (* The types [tyvar] and [ty] of decoded type variables and decoded types
     are provided by the module [O]. *)
  open O

  (* ---------------------------------------------------------------------- *)

  (* The type [range] describes a range in the source code.
     It is not interpreted by the solver. It is used only
     as part of error reports. *)
  type range =
    Lexing.position * Lexing.position

  (* The type [variable] describes the solver's type variables. *)
  type variable

  (* [fresh t] creates a fresh type variable. *)
  val fresh: unit -> variable

  (* ---------------------------------------------------------------------- *)

  (* The syntax of constraints is as follows. *)
  type _ rawco =

    (* Range information. *)
  | CRange : range * 'a rawco -> 'a rawco

    (* Truth. *)
  | CTrue : unit rawco

    (* A conjunction of two constraints. *)
  | CConj : 'a rawco * 'b rawco -> ('a * 'b) rawco

    (* An equation between two types, represented by two unifier variables,
       [v1] and [v2]. These variables must have been correctly bound via
       [CExist] or by [CLet]. *)
  | CEq : variable * variable -> unit rawco

    (* An existentially quantified constraint, [exists (v=s).C]; [v] is
       a syntactic variable, and it is optionally constrained to be equal to
       some structure [s]. *)
  | CExist : variable * variable structure option * 'a rawco -> 'a rawco

  | CWitness : variable -> ty rawco

    (* An application of a constraint abstraction, [x w]. The term
       variable refers to a constraint abstraction (in other words, a type
       scheme) that has been defined earlier via [CDef] or [CLet]. This
       abstraction is applied to the type variable [w] (in other words, [w] must
       be an instance of this type scheme). The solver will compute and return
       a list of witnesses, indicating how the type scheme was instantiated. *)
  | CInstance : tevar * variable -> ty list rawco

    (* A trivial type scheme definition, [def x = v in C]. This binds [x] to
       the monomorphic type scheme [v] in the constraint [C]. In other words,
       an instantiation constraint [x w] will be satisfied by unifying [v] and
       [w]. *)
  | CDef : tevar * variable * 'a rawco -> 'a rawco

    (* A nontrivial type scheme definition, [let [x, v]* C1 in C2 [vs]].
       In short, for each pair [x, v] in the list, the term variable [x]
       is bound to the constraint abstraction [\v.C1]. The environment is extended
       with all of these new bindings when [C2] is examined.

       In addition to the result of solving [C1] and [C2], the solver returns
       a list of type variables that must be universally quantified in
       the left-hand side of the [let] construct so as to be in scope
       when [C1] is decoded, and the list of inferred type schemes that
       represent the simplified forms of the abstractions [\v.C1]. *)
  | CLet : (tevar * variable) list
           * 'a rawco
           * 'b rawco
         -> (tyvar list * (tevar * scheme) list * 'a * 'b) rawco

  | CMap : 'a rawco * ('a -> 'b) -> 'b rawco

  val print_rawco : Format.formatter -> _ rawco -> unit

  (* ---------------------------------------------------------------------- *)

  (* The function [solve] is parameterized by the flag [rectypes], which
     indicates whether recursive types are permitted. It expects a constraint
     and solves it; that is, either it fails with an exception, or it succeeds
     and returns a mapping from constraint variables to inferred types. *)

  exception Unbound of range * tevar
  exception Unify of range * O.ty * O.ty
  exception Cycle of range * O.ty

  val solve: rectypes:bool -> 'a rawco -> 'a
end
